from cubelang.sides import CubeSide, ICubeSide, CubeSideView, Component
from cubelang.orientation import Color

import pytest


def side_to_string(side: ICubeSide):
    out = []
    for i in range(side.rows):
        for j in range(side.columns):
            color = side.colors[i, j].name[0].upper()
            out.append(color)
        out.append("/")
    return "".join(out[:-1])


def create_side(rows: int, columns: int, items: str):
    colors = {
        "Y": Color.YELLOW,
        "W": Color.WHITE,
        "R": Color.RED,
        "O": Color.ORANGE,
        "B": Color.BLUE,
        "G": Color.GREEN
    }

    side = CubeSide(rows, columns, Color.RED)
    for i, row in enumerate(items.split("/")):
        for j, char in enumerate(row):
            side.colors[i, j] = colors[char]
    return side


@pytest.mark.parametrize("rotation, expected, rows, columns", [
    [-1, "RO/GB/YW", 3, 2],
    [0, "RO/GB/YW", 3, 2],
    [1, "OBW/RGY", 2, 3],
    [2, "WY/BG/OR", 3, 2],
    [3, "YGR/WBO", 2, 3],
])
def test_get_set(rotation: int, expected: str, rows: int, columns: int) -> None:
    side = create_side(3, 2, "RO/GB/YW")
    if rotation >= 0:
        side = CubeSideView(side, rotation)
    assert side.rows == rows
    assert side.columns == columns
    assert side_to_string(side) == expected


def test_rotation_illegal():
    with pytest.raises(Exception):
        side = create_side(2, 3, "YWB/OBG")
        side.rotate(1)


@pytest.mark.parametrize("width, height, items, amount, expected", [
    [2, 3, "YWR/OBG", 2, "GBO/RWY"],
    [3, 3, "GBO/RWY/OWB", 1, "ORG/WWB/BYO"],
    [3, 3, "GBO/RWY/OWB", 2, "BWO/YWR/OBG"],
    [3, 3, "GBO/RWY/OWB", 3, "OYB/BWW/GRO"],
    [4, 4, "YOBG/RWYR/OBBY/GBRW", 3, "GRYW/BYBR/OWBB/YROG"],
    [4, 4, "YOBG/RWYR/OBBY/GBRW", 1, "GORY/BBWO/RBYB/WYRG"]
])
def test_rotation(width: int, height: int, items: str, amount: int, expected: str):
    side = create_side(width, height, items)
    side.rotate(amount)
    assert side_to_string(side) == expected


def test_get_row():
    side = create_side(4, 4, "YOBG/RWYR/OBBY/GBRW")
    assert (list(map(lambda x: x.color, side.get_row(1))) ==
            [Color.RED, Color.WHITE, Color.YELLOW, Color.RED])


def test_get_column():
    side = create_side(4, 4, "YOBG/RWYR/OBBY/GBRW")
    assert (list(map(lambda x: x.color, side.get_column(1))) ==
            [Color.ORANGE, Color.WHITE, Color.BLUE, Color.BLUE])


def test_set_row():
    side = create_side(4, 4, "YOBG/RWYR/OBBY/GBRW")
    side.set_row(1, list(map(lambda c: Component[None](c, None),
                             [Color.BLUE, Color.RED, Color.GREEN, Color.YELLOW])))
    assert side_to_string(side) == "YOBG/BRGY/OBBY/GBRW"


def test_set_column():
    side = create_side(4, 4, "YOBG/RWYR/OBBY/GBRW")
    side.set_column(1, list(map(lambda c: Component[None](c, None),
                                [Color.BLUE, Color.RED, Color.GREEN, Color.YELLOW])))
    assert side_to_string(side) == "YBBG/RRYR/OGBY/GYRW"



from cubelang.compiler.code_map import CodeMap
from cubelang.compiler import Stack, parser
from cubelang.execution import ExecutionContext
from cubelang.compiler.types import Function, Integer, Void, Color
from cubelang.cube_runtime import CubeRuntime
from cubelang.stdlib import stdlib
from cubelang.cube import Cube
from cubelang import orientation
from cubelang.orientation import Orientation
from cubelang.execution.executor import ITracebackWriter

from unittest.mock import MagicMock


class MockTracebackWriter(ITracebackWriter):
    def print_traceback(self, error: RuntimeError, code_map: CodeMap) -> None:
        raise RuntimeError()


def test_flip_flops():
    code = """
        let count: int
        while top[1, 1] != top[2, 2] or count == 0 do 
            RUR'U'
            count = count + 1
        end
        out(count)
    """

    out_fn = MagicMock()

    stack = Stack()
    finish_function = MagicMock()
    cube_runtime = CubeRuntime(Cube((3, 3, 3)), Orientation(), lambda action: None, finish_function)
    cube_runtime.functions.initialize_stack(stack)
    stdlib.initialize_stack(stack)
    stack.add_global("out", Function(([Integer], Void)))

    globals = {"out": out_fn, **stdlib.exec_globals, **cube_runtime.functions.exec_globals}

    executor = ExecutionContext(globals)
    executor.compile(parser.parse(code, stack))
    executor.execute(MockTracebackWriter())
    cube_runtime.finished()

    out_fn.assert_called_once_with(6)
    finish_function.assert_called_once()


def test_orient():
    code = """
        orient top: {G--/---/---}, bottom: {--Y/---/---} then
            out(red)
        else-orient top: {-W-/---/---}, right: {---/---/-O-} then
            out(top[1, 1])
        end
    """

    out_fn = MagicMock()
    stack = Stack()
    cube_runtime = CubeRuntime(Cube((3, 3, 3)), Orientation(), lambda action: None, lambda: None)
    cube_runtime.functions.initialize_stack(stack)
    stdlib.initialize_stack(stack)
    stack.add_global("out", Function(([Color], Void)))

    globals = {"out": out_fn, **stdlib.exec_globals, **cube_runtime.functions.exec_globals}

    executor = ExecutionContext(globals)
    executor.compile(parser.parse(code, stack))
    executor.execute(MockTracebackWriter())
    cube_runtime.finished()
    out_fn.assert_called_once_with(orientation.Color.WHITE)



from argparse import ArgumentParser, Namespace

from ..postprocessing import OptimizingPostprocessor, FormattingPostprocessor, \
    OrientationFreezePostprocessor, chain, PostprocessorBase


def init_postprocessors_args_parser(parser: ArgumentParser) -> None:
    parser.add_argument("-o", "--not-optimize", dest="optimize", action='store_false',
                        help="do not eliminate redundant turns and rotations")
    parser.add_argument("-r", "--no-rotations", dest="freeze_orientation", action="store_true",
                        help="do not produce rotations")


def print_action(action: str):
    print(action, end="")


def build_postprocessors_chain(args: Namespace) -> PostprocessorBase:
    postprocessors = []

    if args.freeze_orientation:
        postprocessors.append(OrientationFreezePostprocessor())
    if args.optimize:
        postprocessors.append(OptimizingPostprocessor())
    postprocessors.append(FormattingPostprocessor())

    chain(*postprocessors, print_action)
    return postprocessors[0]



from types import CodeType
from typing import Iterator, Dict, Any
from ..compiler.expression import Expression
from ..compiler.codeio import CodeStream
from ..compiler.stack import VariablesPool
from ..compiler.code_map import CodeMap
from .rt_error import RuntimeError, TerminateExecutionError
from .rt_function import runtime_function
from abc import ABC, abstractmethod


class ITracebackWriter(ABC):
    @abstractmethod
    def print_traceback(self, error: RuntimeError, code_map: CodeMap) -> None:
        pass


class ExecutionContext:
    def __init__(self, globals: Dict[str, Any]):
        self.source: CodeType = None
        self.globals: Dict[str, Any] = globals
        self.globals["runtime_function"] = runtime_function
        self.code_map = CodeMap()

    def compile(self, program: Iterator[Expression]):
        source = self.compile_source(program)
        # print(source)
        self.source = compile(source, "<string>", "exec")

    def compile_source(self, program: Iterator[Expression]) -> str:
        stream = CodeStream()
        variables = VariablesPool()

        for expression in program:
            expression.generate(variables, stream, self.code_map)
        return stream.get_contents()

    def execute(self, error: ITracebackWriter) -> bool:
        if self.source is None:
            raise RuntimeError("Illegal state: program must be compiled first")
        try:
            exec(self.source, self.globals)
            return True
        except TerminateExecutionError:
            return True
        except Exception as e:
            e = RuntimeError.update_error(None, e)
            error.print_traceback(e, self.code_map)
            return False
